from scipy.interpolate import CubicSpline  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import ArrayLike, NDArray
from numpy import (
    append,
    cos,
    float64,
    linspace,
    array,
    logical_and,
    logical_or,
    sin,
    where,
    arange,
    zeros,
)
from scipy.signal import detrend, welch
from scipy.fft import rfftfreq, rfft, irfft
from science_signal import interpolate


class Signal:
    """
    A signal with amplitude linked to times/frequencies
    """

    def __init__(
        self, x: NDArray[float64], y: NDArray[float64]
    ) -> None:
        if len(x) != len(y):
            raise ValueError(
                _(
                    "each value of a signal must be linked with a "
                    + "frequency/time"
                )
            )
        self.x = x
        self.y = y
        self.index = -1
        try:
            self.sampling: float = self.x[1] - self.x[0]
        except IndexError:
            raise ValueError(
                _("the signal must contains more than one value")
            )
        try:
            self.rate = 1 / self.sampling
        except ZeroDivisionError:
            raise ValueError(
                _(
                    "each value of the signal should be linked with a "
                    + "single frequency/time"
                )
            )

    def spline(self) -> CubicSpline:
        """
        Return the x to y spline of this signal
        """
        return CubicSpline(self.x, self.y)

    def cos(self) -> "Signal":
        """
        Return the cosinus of this signal
        """
        return Signal(
            self.x,
            cos(self.y),
        )

    def cut(self, start: None | float = None, end: None | float = None) -> "Signal":
        """
        Cut signal from a start x to an end x
        """
        if start is None:
            start = min(self.x)
        if end is None:
            end = max(self.x)
        indexes = where(logical_and(self.x >= start, self.x <= end))
        return Signal(
            self.x[indexes],
            self.y[indexes],
        )

    def detrend(self, type: str = "linear") -> "Signal":
        """
        Short alias for scipy.detrend with default value
        """
        return Signal(
            self.x,
            detrend(self.y, type=type),
        )

    def extend(
        self, start: None | float = None, end: None | float = None
    ) -> "Signal":
        """
        Extend a signal to a start x or an end x
        """
        x = self.x
        y = self.y
        if start is not None:
            x = append(arange(start, x[0], 1 / self.rate), x)
            y = append(zeros(len(x) - len(y)), y)
        if end is not None:
            x = append(x, arange(x[-1], end, 1 / self.rate))
            y = append(y, zeros(len(x) - len(y)))
        return Signal(
            x,
            y,
        )

    def filter(
        self, start: None | float = None, end: None | float = None
    ) -> "Signal":
        freq_x = rfftfreq(len(self), self.sampling)
        freq_y = rfft(self.y)

        if start is None:
            start = min(abs(freq_x))
        if end is None:
            end = max(abs(freq_x))

        index_to_remove = where(
            logical_or(abs(freq_x) <= start, abs(freq_x) >= end)
        )
        freq_y[index_to_remove] = 0

        y = irfft(freq_y)
        return Signal(
            self.x[: len(y)],
            y[: len(self.x)],
        )

    def psd(self, fft_length: int = 10) -> "Signal":
        """
        Compute psd of a given signal
        """
        freq, psd = welch(
            self.y,
            self.rate,
            nperseg=fft_length * self.rate,
        )

        return Signal(freq, psd)

    def sin(self) -> "Signal":
        """
        Return the sinus of this signal
        """
        return Signal(self.x, sin(self.y))

    def sqrt(self) -> "Signal":
        """
        Return the root square of this signal
        """
        return Signal(self.x, self.y ** (1 / 2))

    def operator_signal(
        self, other: "Signal", operator: str
    ) -> "Signal":
        """
        Apply an operator to this signal with another one
        """
        try:
            if (self.x != other.x).any():  # fail if different length
                raise ValueError(
                    "not an error"
                )  # fail if different value
        except ValueError:
            self, other = interpolate(self, other)
        if operator == "+":
            return Signal(self.x, self.y + other.y)
        if operator == "-":
            return Signal(self.x, self.y - other.y)
        if operator == "*":
            return Signal(self.x, self.y * other.y)
        if operator == "/":
            return Signal(self.x, self.y / other.y)

        raise Exception(_("unknown operator {}").format(operator))

    def operator_arraylike(
        self, other: ArrayLike, operator: str
    ) -> "Signal":
        """
        Apply an operator to this signal (signal operator other)
        """
        try:
            if len(self) != len(other):  # fail if single value
                # suppose same range but different rates
                return self.operator_signal(
                    Signal(
                        linspace(self.x[0], self.x[-1], len(others)),
                        array(other),
                    ),
                    operator,
                )
        except TypeError:
            pass
        if operator == "+":
            return Signal(self.x, self.y + other)
        if operator == "-":
            return Signal(self.x, self.y - other)
        if operator == "*":
            return Signal(self.x, self.y * other)
        if operator == "/":
            return Signal(self.x, self.y / other)
        raise Exception(_("unkown operator {}".format(operator)))

    def roperator_arraylike(
        self, other: ArrayLike, operator: str
    ) -> "Signal":
        """
        Apply an operator to this signal (other operator signal)
        """
        try:
            if len(self) != len(other):  # fail if single value
                return Signal(
                    linspace(self.x[0], self[-1], len(other)),
                    array(other),
                ).operator_signal(self, operator)
        except TypeError:
            pass
        if operator == "-":
            return Signal(self.x, other - self.y)
        if operator == "/":
            return Signal(self.x, other / self.y)
        raise Exception(_("unknown operator {}".format(operator)))

    def __add__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "+")
        return self.operator_arraylike(other, "+")

    def __radd__(self, other: ArrayLike | "Signal") -> "Signal":
        return self + other  # symetric

    def __sub__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "-")
        return self.operator_arraylike(other, "-")

    def __rsub__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return other.operator_signal(self, "-")
        return self.roperator_arraylike(other, "-")

    def __mul__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "*")
        return self.operator_arraylike(other, "*")

    def __rmul__(self, other: ArrayLike | "Signal") -> "Signal":
        return self * other

    def __truediv__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "/")
        return self.operator_arraylike(other, "/")

    def __rtruediv__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return other.operator_signal(self, "/")
        return self.roperator_arraylike(other, "/")

    def __pow__(self, other: float | int) -> "Signal":
        return Signal(self.x, self.y**other)

    def __array__(self) -> NDArray[float64]:
        return array([self.x, self.y])

    def __abs__(self) -> "Signal":
        return Signal(self.x, abs(self.y))

    def __neg__(self) -> "Signal":
        return Signal(self.x, -self.y)

    def __len__(self) -> int:
        return len(self.x)

    def __next__(self) -> tuple[float, float]:
        self.index += 1
        if self.index == len(self):
            raise StopIteration
        return self.x[self.index], self.y[self.index]

    def __getitem__(self, key: int) -> tuple[float, float]:
        return self.x[key], self.y[key]

    def __iter__(self) -> "Signal":
        self.index = -1
        return self
